// speechcaliper.c
// for NerdKits with ATmega328P
// http://www.nerdkits.com/

#define F_CPU 14745600

#include <stdio.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <inttypes.h>

#include "../libnerdkits/delay.h"
#include "../libnerdkits/uart.h"

#include "huffcode.h"

#include "sound/headers/sound_struct.h"

#include "sound/headers/one.h"
#include "sound/headers/two.h"
#include "sound/headers/three.h"
#include "sound/headers/four.h"
#include "sound/headers/five.h"
#include "sound/headers/six.h"
#include "sound/headers/seven.h"
#include "sound/headers/eight.h"
#include "sound/headers/nine.h"
#include "sound/headers/zero.h"
#include "sound/headers/minus.h"
#include "sound/headers/point.h"
#include "sound/headers/inches.h"
#include "sound/headers/millimeters.h"

// PIN DEFINITIONS
//
// PD3 - audio out (OC2B, pin 5)
//
// PB1 -- (inverted) data
// PB2 -- (inverted) clock

////////////////////////////////////////////////////////////////////

uint8_t multi_sample_pin(uint8_t num_samples, volatile uint8_t *port, uint8_t pin) {
  // sample a bunch of times and take the best guess
  // num_samples should be an odd number (maybe 5 or 9)
  // call this as multi_sample_pin(5, &PINC, PC0)
  uint8_t samples_0 = 0;
  uint8_t samples_1 = 0;
  uint8_t i;
  for(i=0; i<num_samples; i++) {
    if((*port & (1<<(pin))) == 0) {
      samples_0++;
    } else {
      samples_1++;
    }
  }
  
  if(samples_0 > samples_1) {
    return 0;
  } else {
    return 1;
  }
}

////////////////////////////////////////////////////////////////////

int32_t caliper_value = 0;
uint8_t caliper_mode = 0;
#define CALIPER_MODE_MM 0
#define CALIPER_MODE_INCH 1

void calipers_setup() {
  // PB1, PB2 inputs with pull-up enabled
  DDRB &= ~(1<<PB1);
  PORTB |= (1<<PB1);

  DDRB &= ~(1<<PB2);
  PORTB |= (1<<PB2);
}

void wait_for_quiet() {
  // wait for at least 2ms of quiet (low, logical 1) on the /CLK line
  uint16_t quiet_count = 0;
  uint8_t data;
  while(quiet_count < 2000) {
    quiet_count++;
    data = PINB & (1<<PB2);
    if(data) quiet_count = 0;
  }
}

void read_bits() {
  // wait for rising /CLK edge
  // then sample /DATA on the /CLK falling edges until we have 24 bits

  int32_t bitval = 1;
  int32_t accum = 0;
  uint8_t data_bit = 0;
  uint8_t i = 0;
  uint8_t mode = 0;

  // wait for rising /CLK edge (logical 0 -> 1)
  while(!multi_sample_pin(5, &PINB, PB2) == 1) { }
  
  // read 24 bits
  for(i=0; i < 24; i++) {
    // wait for /CLK falling edge (logical 1 -> 0)
    while(!multi_sample_pin(5, &PINB, PB2) == 0) { }
    
    // sample data
    data_bit = !multi_sample_pin(5, &PINB, PB1);
    
    // store it
    if(i < 20) {
      // data bit
      if(data_bit) {
        // logical 1 -- add to accumulator
        accum |= bitval;
      } else {
        // logical 0 -- do nothing
      }
    } else if(i==20) {
      // sign bit
      if(data_bit) {
        // logical 1 -- negative sign showing
        accum = -accum;
      }
      // publish value
      caliper_value = accum;
    } else if(i==23) {
      // mode bit
      if(data_bit) {
        // logical 1 -- inch mode, 0.0005 inches per count
        mode = CALIPER_MODE_INCH;
      } else {
        // logical 0 -- millimeter mode, 0.01 mm per count
        mode = CALIPER_MODE_MM;
      }
      // publish mode
      caliper_mode = mode;
    }
    
    // move over the next bit mask
    bitval <<= 1; 

    // wait for rising /CLK edge (logical 0 -> 1)
    while(!multi_sample_pin(5, &PINB, PB2) == 1) { }
  }
}

////////////////////////////////////////////////////////////////////

volatile uint8_t pwm_cycle_counter;

ISR(TIMER2_OVF_vect) {
  pwm_cycle_counter++;
}

void sound_setup() {
  // Timer2, OC2B enable, fast PWM, CLK/1
  TCCR2A = (1<<COM2B1) | (1<<WGM21) | (1<<WGM20);
  TCCR2B = (1<<CS20);
  // interrupt on overflow
  TIMSK2 = (1<<TOIE2);
  // PD3 output
  DDRD |= (1<<PD3);
}

PGM_P cur_byte_ptr;
uint8_t cur_byte;
uint8_t cur_bit_idx;

uint8_t get_next_bit() {
  uint8_t rv = 0;
  if((cur_byte & (1<<cur_bit_idx)) != 0) {
    rv = 1;
  }
  cur_bit_idx++;
  if(cur_bit_idx == 8) {
    cur_byte_ptr++;
    cur_byte = pgm_read_byte(cur_byte_ptr);
    cur_bit_idx = 0;
  }  
  return rv;
}

void play(sound_t *mysound) {
  uint16_t total_samples = 0;
  uint16_t samples_read = 0;

  // keep track of sample count
  total_samples = mysound->samples;

  // cue beginning
  cur_byte_ptr = mysound->data;
  cur_byte = pgm_read_byte(cur_byte_ptr);
  cur_bit_idx = 0;


  printf_P(PSTR("play(): samples = %d\r\n"), total_samples);
  
  int16_t next_sym;
  int16_t y=0, y_1=0, y_2=0;
  int16_t y_out;
  uint8_t y_out_real;
  
  while(samples_read < total_samples) {
    next_sym = get_next_symbol();
    //printf_P(PSTR("%d\r\n"), next_sym);
    samples_read++;
    
    // compute it
    y = (next_sym) + (2*y_1) + (-y_2);
    
    // clip output
    y_out = y + 128;
    if(y_out < 0) {
      y_out_real = 0;
    } else if(y_out > 255) {
      y_out_real = 255;
    } else {
      y_out_real = y_out;
    }
    
    // wait until next period
    while(pwm_cycle_counter < 7) {
      sleep_enable();
      sleep_cpu();
      sleep_disable();
    }
    pwm_cycle_counter = 0;
    
    // set output
    OCR2B = y_out_real;
    
    // shift
    y_2 = y_1;
    y_1 = y;
  }
}

////////////////////////////////////////////////////////////////////

void play_digit(uint8_t digit) {
  // play a digit, 0 through 9
  switch(digit) {
    case 0: play(&sound_zero); return;
    case 1: play(&sound_one); return;
    case 2: play(&sound_two); return;
    case 3: play(&sound_three); return;
    case 4: play(&sound_four); return;
    case 5: play(&sound_five); return;
    case 6: play(&sound_six); return;
    case 7: play(&sound_seven); return;
    case 8: play(&sound_eight); return;
    case 9: play(&sound_nine); return;
  }
}

void play_integer(uint32_t val, uint8_t min_digits) {
  // play a positive integer, 0 to 4,294,967,296
  // if min_digits > 0, then say leading zeros up to X digits.
  uint8_t is_started = 0;

  uint8_t i = 10;
  uint32_t div = 1000000000;
  while(div > 0) {
    if( (val >= div) || (min_digits >= i) || (is_started) ) {
      play_digit( val / div );
      val = val % div;
      is_started = 1;
    }
    div = div / 10;
    i--;
  }
}

void play_decimal(int32_t val, uint8_t decimal_digits) {
  // if val=5321 and decimal_digits=2, then say "53.21"
  
  // trap negative numbers
  if(val < 0) {
    play(&sound_minus);
    play_decimal(-val, decimal_digits);
    return;
  }
  
  // make divisor for decimalness
  int32_t ddd = 1;
  uint8_t i;
  for(i=0; i<decimal_digits; i++) {
    ddd = ddd * 10;
  }
  
  
  // separate sides of the decimal point
  int32_t p_right = val % ddd;
  int32_t p_left = val / ddd;
  
  // say left side
  // and at least say "0"
  play_integer(p_left, 1); 
  
  // if there's a right side, say that too
  if(decimal_digits > 0) {
    play(&sound_point);
    play_integer(p_right, decimal_digits);
  }
}

////////////////////////////////////////////////////////////////////

#define ABS(x)  ( ((x) >= 0) ? (x) : (-(x)) )

int main() {
  // power reduction
  // shut down TWI, SPI, ADC modules
  PRR |= (1<<PRTWI) | (1<<PRSPI) | (1<<PRADC);
  // shut down Timer1
  PRR |= (1<<PRTIM1);
  // shut down Timer0
  PRR |= (1<<PRTIM0);

  // inits
  calipers_setup();
  sound_setup();
  
  // start up the serial port
  uart_init();
  FILE uart_stream = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);
  stdin = stdout = &uart_stream;

  // enable interrupts
  sei();

  uint8_t nomovecutoff = 0;
  int32_t tmpr;
  int32_t tmpr_1=0, tmpr_2=0, tmpr_3=0;
  int32_t res_inch; // 1000 = 1 inch
  int32_t res_mm;   //  100 = 1 mm
  while(1) {
    // read from calipers()
    wait_for_quiet();
    read_bits();
    
    // print to serial port
    printf_P(PSTR("%ld\t%d\r\n"), caliper_value, caliper_mode);
    
    // convert to units of milli-inches
    res_inch = caliper_value / 2;
    
    // convert to units of 100ths-of-millimeters
    res_mm = caliper_value;
    
    // non-movement criteria
    if(caliper_mode == CALIPER_MODE_INCH) {
      nomovecutoff = 4; // 0.002 inches
    } else if(caliper_mode == CALIPER_MODE_MM) {
      nomovecutoff = 5; // 0.05 mm
    }
    
    // is it similar to the last one?
    tmpr = caliper_value;
    if( (ABS(tmpr - tmpr_1) < nomovecutoff) && (ABS(tmpr - tmpr_2) < nomovecutoff) && (ABS(tmpr - tmpr_3) < nomovecutoff) ) {
      // say out loud!!
      if(caliper_mode == CALIPER_MODE_INCH) {
        play_decimal(res_inch, 3);
        play(&sound_inches);
      } else if(caliper_mode == CALIPER_MODE_MM) {
        play_decimal(res_mm, 2);
        play(&sound_millimeters);
      }
      // enforce a small delay before we say it again
      tmpr = tmpr - 500;
    }
    
    // store for comparison
    tmpr_3 = tmpr_2;
    tmpr_2 = tmpr_1;
    tmpr_1 = tmpr;  
  }

  return 0;
}

